import path, { dirname } from "node:path";
import { Glob } from "bun";
import { FileNotFoundError, MuseError } from "./errors";
import { parseMuseFile } from "./parse";
import {
  type Binding,
  BindingSchema,
  type MuseEntry,
  type MuseProcessor,
} from "./types";
import { resolveFilePath } from "./util";

async function loadProcessor(
  bindingDirname: string,
  processorPath: string,
): Promise<MuseProcessor> {
  const resolvedProcessorPath = path.resolve(bindingDirname, processorPath);
  const { process, name, description } = await import(resolvedProcessorPath);

  if (!process || typeof process !== "function") {
    throw new MuseError(
      `Processor at ${processorPath} does not export a process() function`,
    );
  }

  const processorName = String(name ?? "Unnamed Processor");
  const processorDescription = String(description ?? "No description provided");

  return {
    filePath: resolvedProcessorPath,
    process: process as MuseProcessor["process"],
    name: processorName,
    description: processorDescription,
  };
}
/**
 * Given a path, find the binding file
 * If the path is to a directory, check for a binding inside
 * Check order: binding.json, binding.yaml, binding.toml, binding.md
 * Otherwise throw an error
 *
 * @param bindingPath - The path to the binding file
 * @returns The absolute resolved path to the binding file
 */
export async function resolveBindingPath(bindingPath: string): Promise<string> {
  // If the path does not specify a filename, use binding.yaml
  const inputLocation = Bun.file(bindingPath);

  // If it is a directory, try for the four main supported types
  const stat = await inputLocation.stat();
  if (stat.isDirectory()) {
    const jsonPath = await resolveFilePath(
      path.resolve(bindingPath, "binding.json"),
    );
    if (jsonPath) {
      return jsonPath;
    }

    const yamlPath = await resolveFilePath(
      path.resolve(bindingPath, "binding.yaml"),
    );
    if (yamlPath) {
      return yamlPath;
    }

    const tomlPath = await resolveFilePath(
      path.resolve(bindingPath, "binding.toml"),
    );
    if (tomlPath) {
      return tomlPath;
    }

    const mdPath = await resolveFilePath(
      path.resolve(bindingPath, "binding.md"),
    );
    if (mdPath) {
      return mdPath;
    }

    throw new FileNotFoundError(bindingPath);
  }

  const exists = await inputLocation.exists();

  if (!exists) {
    throw new FileNotFoundError(bindingPath);
  }

  // If it is a file, return the path
  return path.resolve(bindingPath);
}

export async function loadBinding(initialPath: string): Promise<Binding> {
  // Resolve the file location if possible
  const bindingPath = await resolveBindingPath(initialPath);
  const bindingDirname = dirname(bindingPath);
  const loadedBindingData = await parseMuseFile(bindingPath);

  if (loadedBindingData.length === 0) {
    throw new MuseError("No entries found in binding file");
  }

  // Load and parse the Binding YAML content
  const parsedBinding = BindingSchema.parse(loadedBindingData[0].data);

  // Prepare collections to load into
  const extendsFrom: Binding[] = [];
  const entries: MuseEntry[] = [];

  // Process imported bindings and bring in their entries
  for (const includePath of parsedBinding.include) {
    const pathFromHere = path.resolve(bindingDirname, includePath);
    const extendBindingPath = await resolveBindingPath(pathFromHere);
    const loadedBinding = await loadBinding(extendBindingPath);
    extendsFrom.push(loadedBinding);
    entries.push(...loadedBinding.entries);
  }

  // Aggregate file paths to import
  const includedFilePaths = parsedBinding.files
    .flatMap((thisGlob) => {
      return Array.from(
        new Glob(thisGlob).scanSync({
          cwd: bindingDirname,
          absolute: true,
          followSymlinks: true,
          onlyFiles: true,
        }),
      );
    })
    .filter((filePath) => filePath !== bindingPath);

  // Load files
  const loadedEntries = await Promise.all(
    includedFilePaths.map((filePath) => {
      return parseMuseFile(filePath, {
        contentKey: parsedBinding.contentKey,
      });
    }),
  );

  // Add loaded entries to main list
  entries.push(...loadedEntries.flat());

  // Load and check processors
  const processors: MuseProcessor[] = await Promise.all(
    parsedBinding.processors.map(loadProcessor.bind(null, bindingDirname)),
  );

  return {
    _raw: parsedBinding,
    bindingPath,
    contentKey: parsedBinding.contentKey,
    entries,
    options: parsedBinding.options,
    processors: processors,
    imports: extendsFrom,
    meta: {},
  };
}
